
##
## Reference implementation of the (1+1) Limited Memory Matrix Adaptation
## Evolution Strategy; (1+1)-LM-MA-ES.
##
## Copyright (C) 2024 Tobias Glasmachers
## 
## Permission is hereby granted, free of charge, to any person obtaining a
## copy of this software and associated documentation files (the
## "Software"), to deal in the Software without restriction, including
## without limitation the rights to use, copy, modify, merge, publish,
## distribute, sublicense, and/or sell copies of the Software, and to
## permit persons to whom the Software is furnished to do so, subject to
## the following conditions:
##
## The above copyright notice and this permission notice shall be included
## in all copies or substantial portions of the Software.
##
## THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
## OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
## MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
## IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
## CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
## TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
## SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
##


import numpy as np
import sys
import time


def lm_ma_es_1p1(x, sigma, f, f_target, budget, verbosity = 1):
	start = time.time()

	# preparation
	n = len(x)                                 # search space dimensionality
	k = int(4 + np.floor(3 * np.log(n)))       # number of directions in the covariance model
	alpha = 2 / n                              # step size adaptation learning rate
	c_del = np.zeros((k,))                     # information deletion learning rate
	c_new = np.zeros((k,))                     # information update learning rate
	for i in range(k):
		c_del[i] = 1 / (4.0**i * n)
		c_new[i] = np.sqrt(2) / n * 6**(-i/2)
	c_keep = (1 - c_del).reshape(k, 1)         # information to keep
	m = np.zeros((k, n))                       # k path vectors (covariance information)

	if verbosity >= 1:
		print("[1+1 LM-MA-ES]  f=" + f.__name__ + "  n=" + str(n) + "  k=" + str(k) + "  target=" + str(f_target))

	filename = "logs/" \
			+ sys.argv[1] \
			+ "_" + sys.argv[2] \
			+ "_" + sys.argv[3] \
			+ ".csv"
	logfile = open(filename, "w+")

	prev_report = 0
	prev_fitness = 1e100
	def report(fe, fitness, sigma, force):
		nonlocal prev_report, prev_fitness, logfile
		t = time.time()
		if force or t - prev_report >= 0.2:
			sys.stdout.write('\b' * 100);
			sys.stdout.write("    #FE: " + str(fe) + "  fitness: " + str(fitness) + "  step size: " + str(sigma) + "     \b\b\b\b\b")
			sys.stdout.flush()
			prev_report = t
		if force or fitness / prev_fitness < 0.95:
			logfile.write(str(fe) + "," + str(fitness) + "," + str(sigma) + "\n")
			logfile.flush()
			prev_fitness = fitness

	# evaluate parent
	fx = f(x)

	# generation loop
	t = 1   # one individual is already evaluated
	while t < budget:
		# generation counter
		t += 1

		# sample offspring
		z = np.random.randn(n)
		d = z + (m @ z) @ m

		# evaluate offspring
		y = x + sigma * d
		fy = f(y)

		# check for success
		success = (fy <= fx)

		if success:
			# update mean
			x = y
			fx = fy

			# update direction vectors
			m *= c_keep
			m += np.outer(c_new, z)

			# check stopping criterion
			if fx <= f_target:
				report(t, fx, sigma, True)
				print("  target fitness reached (" + str(time.time() - start) + " seconds)")
				return x, t

		# update step size
		sigma *= np.exp((1 if success else -0.25) * alpha)

		if sigma < 1e-13:
			report(t, fx, sigma, True)
			print("  convergence detected (" + str(time.time() - start) + " seconds)")
			return x, t

		# status report
		report(t, fx, sigma, False)

	# we ran out of budget
	report(budget, fx, sigma, True)
	print("  budget exceeded (" + str(time.time() - start) + " seconds)")
	return x, budget


if __name__ == "__main__":
	def sphere(x):
		return np.dot(x, x)

	elli_factors = np.zeros(0)
	def elli(x):
		global elli_factors
		if elli_factors.shape != x.shape:
			dim = x.shape[0]
			elli_factors = np.zeros(dim)
			for i in range(dim):
				elli_factors[i] = np.power(1e-6, i / (dim - 1))
		return np.dot(x, elli_factors * x)

	def cigar(x):
		return x[0] * x[0] + 1e6 * np.dot(x[1:], x[1:])

	def discus(x):
		return 1e6 * x[0] * x[0] + np.dot(x[1:], x[1:])

	def rosen(x):
		n = len(x)
		return np.sum(100.0 * np.square(x[1:n] - np.square(x[0:n-1])) + np.square(x[0:n-1] - np.ones(n-1)))

	diffpow_factors = np.zeros(0)
	def diffpow(x):
		global diffpow_factors
		if diffpow_factors.shape != x.shape:
			dim = x.shape[0]
			diffpow_factors = np.zeros(dim)
			for i in range(dim):
				diffpow_factors[i] = 2.0 + 4.0 * i / (dim - 1.0)
		return np.sum(np.exp(diffpow_factors * np.log(np.fabs(x))))
	funcname = sys.argv[1]
	dim = int(sys.argv[2])
	instance = int(sys.argv[3])
	problem = {"sphere": sphere, "elli": elli, "cigar": cigar, "discus": discus, "rosen": rosen, "diffpow": diffpow}[funcname]
	np.random.seed(1001 + instance)
	lm_ma_es_1p1(np.random.randn(dim), 1 / np.sqrt(dim), problem, 1e-8, 100000 * dim)
