
##
## Reference implementation of the Multi-objective Limited Memory Matrix
## Adaptation Evolution Strategy; MO-LM-MA-ES.
##
## Copyright (C) 2024 Tobias Glasmachers
## 
## Permission is hereby granted, free of charge, to any person obtaining a
## copy of this software and associated documentation files (the
## "Software"), to deal in the Software without restriction, including
## without limitation the rights to use, copy, modify, merge, publish,
## distribute, sublicense, and/or sell copies of the Software, and to
## permit persons to whom the Software is furnished to do so, subject to
## the following conditions:
##
## The above copyright notice and this permission notice shall be included
## in all copies or substantial portions of the Software.
##
## THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
## OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
## MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
## IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
## CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
## TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
## SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
##


import numpy as np
import sys
import time
import random
import os


class ElitistES:
	"""(1+1) LM-MA-ES with an interface suitable for MOO"""
	def __init__(self, x, fx, sigma, m = None):
		self.n = len(x)
		self.k = int(4 + np.floor(3 * np.log(self.n)))
		self.x = x
		self.fx = fx

		c_del = np.zeros(self.k)
		self.c_new = np.zeros(self.k)
		for i in range(self.k):
			c_del[i] = 1 / (4.0**i * self.n)
			self.c_new[i] = np.sqrt(2) / self.n * 6**(-i/2)
		self.c_keep = (1 - c_del).reshape(self.k, 1)

		self.sigma = sigma
		self.m = np.zeros((self.k, self.n)) if m is None else m

		self.c_sigma = 2 / self.n
		self.c_sigma_update = np.sqrt(self.c_sigma * (2 - self.c_sigma))

	def generate(self):
		"""sample an offspring"""
		z = np.random.randn(self.n)
		d = z + (self.m @ z) @ self.m
		y = self.x + self.sigma * d
		return y, z

	def updateMean(self, y, fy):
		"""update mean and direction vectors; do this only in case of success"""
		if y is not None: self.x = y
		if fy is not None: self.fx = fy

	def updateCovariance(self, z):
		"""update direction vectors; do this only in case of success"""
		self.m *= self.c_keep
		self.m += np.outer(self.c_new, z)

	def updateStepSize(self, success):
		"""update step size based on notion of success (1/5 rule)"""
		self.sigma *= np.exp((1 if success else -0.25) * 2 / self.n)


def non_dominated_sorting(y, maxranks = None):
	"""
	Given a set of objective function vectors, return a list of index arrays
	representing the different non-dominance ranks.
	"""
	n = y.shape[0]
	k = y.shape[1]

	# compute full (pairwise) weak dominance relation
	a = np.zeros((n, n, k))
	b = np.zeros((n, n, k))
	a[:] = np.expand_dims(y, 0)
	b[:] = np.expand_dims(y, 1)
	weak_dominance = np.max(a - b, axis=2) <= 0

	# identify non-dominance ranks (starting at 0)
	subset = np.arange(n, dtype=np.uint32)
	ranks = []
	rank = 0
	while n > 0:
		# identify dominated and non-dominated points
		s = np.sum(weak_dominance[:, subset][subset, :], axis=1)
		k = np.min(s)   # should be 1 most of the time
		non_dominated = subset[np.flatnonzero(s <= k)]
		dominated = subset[np.flatnonzero(s > k)]

		assert len(non_dominated) > 0

		# keep the rank
		ranks.append(non_dominated)

		# move to next rank
		rank += 1
		subset = dominated
		n = len(subset)

		if maxranks is not None and rank >= maxranks: break

	return ranks

def hypervolume_contributions(y, boundary = np.inf):
	"""
	Compute contributed hypervolume of all points y, which are assumed to be
	mutually non-dominated. The contribution of boundary points is specified
	by a parameter.
	"""
	n = y.shape[0]
	assert y.shape[1] == 2
	assert n > 0

	order = np.argsort(y[:, 0])
	hv = np.zeros(n)
	hv[order[0]] = boundary
	hv[order[-1]] = boundary
	if n > 2:
		z = y[order]
		assert np.max(z[1:, 0] - z[:-1, 0]) >= 0
		assert np.min(z[1:, 1] - z[:-1, 1]) <= 0
		hv[order[1:-1]] = (z[2:, 0] - z[1:-1, 0]) * (z[:-2, 1] - z[1:-1, 1])
	return hv

def select(y, n, reference):
	"""
	Given a set of objective function vectors, return a vector of n indices.
	Selection is based on non-dominated sorting and hypervolume contributions.
	This function works only for two objectives.
	"""
	assert y.shape[1] == 2
	ndr = non_dominated_sorting(y)
	indices = np.zeros(n, dtype=np.uint32)
	u = 0
	for idx in ndr:
		k = len(idx)
		if u + k <= n:
			indices[u:u+k] = idx
			u += k
			if u == n: break
		else:
			# TODO: this would profit from a more efficient implementation updating HV contributions on removal of a point
			while k > n - u:
				hv = hypervolume_contributions(y[idx])
				i = np.argmin(hv)
				idx[i] = idx[-1]
				idx = idx[:-1]
				k -= 1
			indices[u:] = idx
			break
	return indices

def hypervolume(y, reference):
	"""
	Compute the hypervolume of the set y of objective vectors
	with respect to the given reference point.
	"""
	y = y[np.min(reference - y, axis=1) > 0]   # points dominating the reference point
	if y.shape[0] == 0: return 0
	y = y[non_dominated_sorting(y, 1)[0]]      # non-dominated set
	if y.shape[0] == 0: return 0
	y = y[np.argsort(y[:, 0])]                 # sort the front
	return (reference[0] - y[0, 0]) * (reference[1] - y[0, 1]) + np.sum((reference[0] - y[1:, 0]) * (y[:-1, 1] - y[1:, 1]))

def mo_lm_ma_es(x, sigma, f, reference, target, budget, steadystate = False, verbosity = 1, logfile = None):
	start = time.time()

	# preparation
	mu = x.shape[0]
	es = [ElitistES(x[i], f(x[i]), sigma) for i in range(mu)]
	fe = mu

	if verbosity >= 1:
		print("[MO-LM-MA-ES]  f=" + f.__name__ + "  ref=" + str(reference) + "  mu=" + str(mu) + "  n=" + str(es[0].n) + "  k=" + str(es[0].k) + "  target=" + str(target))

	prev_report = 0
	prev_best = -np.inf
	def report(fe, hypervolume, force=True):
		nonlocal prev_report, prev_best, logfile
		t = time.time()
		if force or t - prev_report >= 0.2:
			sys.stdout.write('\b' * 100);
			sys.stdout.write("    #FE: " + str(fe) + "  hypervolume: " + str(hypervolume) + "     \b\b\b\b\b")
			sys.stdout.flush()
			prev_report = t
		if logfile is not None and hypervolume > prev_best:
			logfile.write(str(fe) + "," + str(hypervolume) + "\n")
			prev_best = hypervolume

	# generation loop
	fx = np.zeros((mu, 2))
	while True:
		if fe >= budget:
			for i in range(mu): fx[i] = es[i].fx
			hv = hypervolume(fx, reference)
			report(budget, hv)
			print("  budget exceeded (" + str(time.time() - start) + " seconds)")
			break

		mutations = []
		if steadystate:
			# sample and evaluate a single offspring
			parent = random.choice(es)
			y, z = parent.generate()
			fy = f(y)
			offspring = ElitistES(y, fy, parent.sigma, m = parent.m)
			es.append(offspring)
			mutations.append((parent, offspring, z))
			fe += 1
		else:
			# sample and evaluate one offspring per parent
			ofs = []
			for parent in es:
				y, z = parent.generate()
				fy = f(y)
				offspring = ElitistES(y, fy, parent.sigma, m = parent.m)
				ofs.append(offspring)
				mutations.append((parent, offspring, z))
			es += ofs
			fe += mu

		# select and check for success
		ids = [id(obj) for obj in es]
		sigmas = [obj.sigma for obj in es]
		fit = np.zeros((len(es), 2))
		for i in range(len(es)): fit[i] = es[i].fx
		idx = select(fit, mu, reference)
		selected = set(idx)

		# updates
		for i in range(len(mutations)):
			p, o, z = mutations[i]
			success = (mu+i) in selected
			if success:
				p.updateCovariance(z)
				o.updateCovariance(z)
			p.updateStepSize(success)
			o.updateStepSize(success)

		# discard truncated individual(s)
		es = [es[i] for i in idx]

		# status report
		for i in range(mu): fx[i] = es[i].fx
		hv = hypervolume(fx, reference)
		report(fe, hv, False)
		if hv >= target:
			print("  target hypervolume reached (" + str(time.time() - start) + " seconds)")
			break

	if logfile is not None: logfile.close()

	# assemble the result
	for i in range(mu):
		x[i] = es[i].x
		fx[i] = es[i].fx
	return x, fx


if __name__ == "__main__":
	# instantiate a problem
	from benchmarks import MOBenchmark
	np.random.seed(1001)
	mu = 20
	dim = 512
	problem_name = "1/I"
	f = MOBenchmark(problem_name, dim, 1)

	# rescale objectives as done in the paper
	f.b1 = 0
	f.b2 = 0
	nadir = f.nadir()
	f.a1 /= nadir[0]
	f.a2 /= nadir[1]

	# optimize
	with open("testrun.txt", "w+") as logfile:
		reference = [10, 10]
		optHV = reference[0]*reference[1] - 0.5 - 0.5 / (mu-1)
		budget = 100 * mu * dim
		pop, fit = mo_lm_ma_es(np.zeros((mu, dim)), 1/np.sqrt(dim), f, reference, optHV - 1e-8, budget, steadystate=False, logfile=logfile)
